
/*
 *  My Application.  
 *  This file shows an example application that saves/restores
 *  QOF objects to/from an SQL database.   Actually, it shows 
 *  three different ways of saving, and one way of restoring.
 *
 */

#include "action.h"
#include "report.h"
#include "interface.h"
#include "duifield-qof.h"
#include "readfile.h"

#include "my-entity.h"

DuiInterface *di = NULL;

/* ============================================================== */
/* Perform various bits of system initialization */
void 
init (void)
{
	qof_object_initialize();
	qof_query_init();
	my_ent_register ();

	di = dui_interface_new();
	dui_interface_parse (di, "my-entity.dui");
	dui_interface_realize (di, FALSE);

	DuiDatabase *db = dui_database_new ("my db object", 
                          // "libdbi", "my-qof-db", 
                          // "odbc", "my-qof-db", 
                          "libpg", "my-qof-db", 
                          // "127.0.0.1", "linas", "abc123");
                          NULL, "linas", NULL);
	dui_interface_add_database (di, db);

	dui_interface_realize (di, FALSE);
}

void 
shutdown(void)
{
	dui_interface_destroy(di);
}

/* ============================================================== */
/* print out the contents of the book. */

static void
print_one (QofEntity *myent, gpointer data)
{
	const char * str = qof_object_printable (MYENT_ID, myent);
	printf ("%s", str);
}

void
print_all (QofBook *book)
{
	qof_object_foreach (MYENT_ID, book, print_one, NULL);
	printf ("-------------------------------------------------\n");
}

/* ============================================================== */
/* 'Copy' an object from the SQL database into a QOF Entity.
 *  This runs the action "hard-coded select" which is a bit
 *  of DUI that includes a hard-coded GUID in it.
 */
void 
copy_from_db(QofBook *book)
{
	/* After reading in the myentity.dui file, there will be a form
	 * named "hard-coded select".  We will want to run that action.
	 * So get its handle now. */
	DuiAction *act;
	act = dui_interface_find_action_by_name (di, "hard-coded select");

	/* To run the report, we need to tell the report what book the object 
	 * instances will live in.  Find the report associated with this
	 * action, and set the book for it. 
	 */
	const char * report_name = dui_action_get_report (act);
	DuiReport *report = dui_interface_find_report_by_name (di, report_name);
	dui_report_add_resolver (report, dui_field_resolve_qof, book);

	/* Cause everything to happen: i.e. issue the SQL Query, and place
	 * the resulting QOF instance into the book */
	dui_action_run (act);

	/* Print all of the entities in the book, so that we can see 
	 * our handiwork. */
	printf ("After running the copy_from_db() routine, "
	        "the contents of the book are:\n");
	print_all (book);
}

/* ----------------------------------------------------------- */
/* 'Copy' an object from the SQL database into a QOF Entity.
 *  This runs the action "kvp select" which obtains the desired
 *  GUID from a value supplied here.
 */
void 
copy_from_db_style_two(QofBook *book)
{
	/* After reading in the myentity.dui file, there will be a form
	 * named "kvp select".  We will want to run that action.
	 * So get its handle now. */
	DuiAction *act;
	act = dui_interface_find_action_by_name (di, "kvp select");

	/* To run the report, we need to tell the report what book the object 
	 * instances will live in.  Find the report associated with this
	 * action, and set the book for it. 
	 */
	const char * report_name = dui_action_get_report (act);
	DuiReport *report = dui_interface_find_report_by_name (di, report_name);
	dui_report_add_resolver (report, dui_field_resolve_qof, book);

	/* Set up the GUID value that will be searched for.  The GUID value
	 * is identified in the action by the KVP key "desired-guid".  So set
	 * that value now */
	dui_interface_kvp_insert (di, "desired-guid", "abcdef0987654321abcdef0987654321");

	/* Cause everything to happen: i.e. issue the SQL Query, and place
	 * the resulting QOF instance into the book */
	dui_action_run (act);

	/* Print all of the entities in the book, so that we can see 
	 * our handiwork. */
	printf ("After running the copy_from_db_style_two() routine, "
	        "the contents of the book are:\n");
	print_all (book);
}

/* ----------------------------------------------------------- */

void
copy_from_db_style_three (QofBook *book)
{
	char query[300];
	DuiDatabase *db;
	DuiDBConnection *db_conn;
	DuiReport *report;
	DuiDBRecordSet *recs = NULL;

	/* Find a database connection.  In this example, we use the connection
	 * that DWI set up for us; although we could also use dui_database_new()
	 * if we wanted to.
	 */
	db = dui_interface_find_database_by_name (di, "my db object");
	db_conn = dui_database_do_realize (db);

	/* Create an 'arbitrary' query.   Of course, the DWI report that will
	 * show the results of this query must be able to understand the 
	 * record (row) that was returned. 
	 */
	sprintf (query, "SELECT * FROM MyEntity WHERE "
	                " myGuid=\'fedcba0123456789abcdef0123456789\';");
	recs = dui_connection_exec(db_conn, query);

	/* Take the returned SQL record, and create a matching QOF object.  
	 * We do this with a named report.  To see the DWI definition of 
	 * this report, grep for 'name="my entity setter"' in the "myentity.dui" 
	 * file.
	 */
	report = dui_interface_find_report_by_name (di, "my entity setter");
	dui_report_set_last_action (report, NULL);
	dui_report_add_resolver (report, dui_field_resolve_qof, book);
	dui_report_show_data (report, recs);
	dui_recordset_free (recs);

	/* Print all of the entities in the book, so that we can see 
	 * our handiwork. */
	printf ("After running the copy_from_db_style_three() routine, "
	        "the contents of the book are:\n");
	print_all (book);
}

/* ============================================================== */
/* 'Copy' a QOF Entity from local memory to the SQL database.
 * Pass in the Guid of the entity that is to be saved to SQL.
 */
void 
copy_to_db(QofBook *book, const GUID *guid)
{
	DuiAction *act;
	act = dui_interface_find_action_by_name (di, "copy-to");

	/* To run the action, we need to tell the action what book the object 
	 * instances will live in.  */
	dui_action_add_resolver (act, dui_field_resolve_qof, book);

	/* Find the field that specifies the matching GUID. 
    * and set the value of out GUID into it.  */
	DuiTxnQuery *qry = dui_action_get_query (act);
	DuiFieldMap *fm = dui_txnquery_get_source_match_term (qry);
	DuiField *fld = &fm->source;
	char buff[GUID_ENCODING_LENGTH+1];
	guid_to_string_buff (guid, buff);
	dui_field_set_const (fld, buff);

	/* Cause everything to happen: i.e. issue the SQL Query, and place
	 * the resulting QOF instance into the book */
	dui_action_run (act);

	/* Print all of the entities in the book, so that we can see 
	 * our handiwork. */
	printf ("After running the copy_to_db() routine, "
	        "the contents of the book are:\n");
	print_all (book);
}

/* ============================================================== */

MyEnt *
add_entity (QofBook *book)
{
	MyEnt *me = my_ent_new (book);

	my_ent_set_a (me, 42);

	time_t now = time(0);
	my_ent_set_b (me, now);

	char buff[300];
	sprintf (buff, "Yo ho ho the time now is %s", ctime(&now));
	int len = strlen (buff);
	buff[len-1] = 0;  /* remove trailing newline char */
	my_ent_set_memo (me, buff);

	printf ("After adding some random entity, "
	        "the contents of the book now are:\n");
	print_all (book);
	return me;
}

int 
main (int argc, char *argv[])
{
	init();

	QofBook *book = qof_book_new();

	copy_from_db (book);
	copy_from_db_style_two (book);
	copy_from_db_style_three (book);

	MyEnt *me = add_entity (book);
	const GUID *guid = qof_entity_get_guid (&me->ent);
	copy_to_db (book, guid);

	shutdown ();

	return 0;
}

/* ======================== END OF FILE ========================= */
/* ============================================================== */
