/********************************************************************\
 * parse-qmap.c -- Plugin for Dui Parser that handle QofMap markup. *
 * Copyright (C) 2004 Linas Vepstas <linas@linas.org>               *
 * http://dwi.sourceforge.net                                       *
 *                                                                  *
 * This library is free software; you can redistribute it and/or    *
 * modify it under the terms of the GNU Lesser General Public       *
 * License as published by the Free Software Foundation; either     *
 * version 2.1 of the License, or (at your option) any later version.
 *                                                                  *
 * This library is distributed in the hope that it will be useful,  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU Lesser General Public License for more details.              *
 *                                                                  *
 * You should have received a copy of the GNU Lesser General Public *
 * License along with this program; if not, contact:                *
 *                                                                  *
 * Free Software Foundation           Voice:  +1-617-542-5942       *
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652       *
 * Boston, MA  02111-1307,  USA       gnu@gnu.org                   *
\********************************************************************/

#include "parse-qmap.h"
#include "qofmap.h"

#include "perr.h"

#include "readfile.h"

/* ============================================================== */
/* The XML Parsing and parse context routines */

struct QmapParseCtxt_s
{
	/** Map being currently parsed */
	QofMap *map;

	/** List of previously parsed maps */
	GList *map_list;
};

/* ============================================================== */

#define CASE(STR,MEMB)                               \
   if (!strcmp (attrs[i], STR)) {                    \
      MEMB = attrs[i+1];                             \
   } else

#define CASE_UNKNOWN                                 \
   {                                                 \
      SYNTAX ("unknown %s attribute %s", eltname, attrs[i]); \
   }

#define REQUIRE2(vara,varb,what)                     \
   if ((NULL == vara) && (NULL == varb))             \
   {                                                 \
      SYNTAX ("%s must specify " what, eltname);     \
      return;                                        \
   }

/* ============================================================== */

static void
qmap_field_start (gpointer self, const char **attrs)
{
	QmapParseCtxt *ctx = self;
	const char * eltname = "<qmap>";
	const char * fieldname=NULL;
	const char * property=NULL;

	int i=0;
	while (attrs[i])
	{
		CASE ("field",       fieldname)
		CASE ("property",    property)
		CASE_UNKNOWN;
		i += 2;
	}
	REQUIRE2(fieldname,property, "An SQL field name and an object property!");

	qof_map_add_field (ctx->map, fieldname, property);
}

static void
qmap_field_end (gpointer self)
{
	// QmapParseCtxt *ctx = self;
}

static DuiParserElementPlugin *
qmap_field_plugin (QmapParseCtxt *ctx)
{
	DuiParserElementPlugin *plg = g_new0 (DuiParserElementPlugin, 1);

	plg->eltname = "qmap";
	plg->start_handler = qmap_field_start;
	plg->end_handler = qmap_field_end;
	plg->user_data = ctx;
	
	return plg;
}

/* ============================================================== */

static void
qmap_match_start (gpointer self, const char **attrs)
{
	QmapParseCtxt *ctx = self;
	const char * eltname = "<qmatch>";
	const char * fieldname=NULL;
	const char * property=NULL;

	int i=0;
	while (attrs[i])
	{
		CASE ("field",       fieldname)
		CASE ("property",    property)
		CASE_UNKNOWN;
		i += 2;
	}
	REQUIRE2(fieldname,property, "An SQL field name and an object property!");

	qof_map_add_match (ctx->map, fieldname, property);
}

static void
qmap_match_end (gpointer self)
{
	// QmapParseCtxt *ctx = self;
}

static DuiParserElementPlugin *
qmap_match_plugin (QmapParseCtxt *ctx)
{
	DuiParserElementPlugin *plg = g_new0 (DuiParserElementPlugin, 1);

	plg->eltname = "qmatch";
	plg->start_handler = qmap_match_start;
	plg->end_handler = qmap_match_end;
	plg->user_data = ctx;
	
	return plg;
}

/* ============================================================== */

static void
qmap_version_start (gpointer self, const char **attrs)
{
	QmapParseCtxt *ctx = self;
	const char * eltname = "<qversion>";
	const char * fieldname=NULL;
	const char * property=NULL;

	int i=0;
	while (attrs[i])
	{
		CASE ("field",       fieldname)
		CASE ("property",    property)
		CASE_UNKNOWN;
		i += 2;
	}
	REQUIRE2(fieldname,property, "An SQL field name and an object property!");

	qof_map_add_version_cmp (ctx->map, fieldname, property);
}

static void
qmap_version_end (gpointer self)
{
	// QmapParseCtxt *ctx = self;
}

static DuiParserElementPlugin *
qmap_version_plugin (QmapParseCtxt *ctx)
{
	DuiParserElementPlugin *plg = g_new0 (DuiParserElementPlugin, 1);

	plg->eltname = "qversion";
	plg->start_handler = qmap_version_start;
	plg->end_handler = qmap_version_end;
	plg->user_data = ctx;
	
	return plg;
}

/* ============================================================== */

static void
qmap_start (gpointer self, const char **attrs)
{
	QmapParseCtxt *ctx = self;
	const char * eltname = "<qof-map>";
	const char * etype=NULL;
	const char * tabname=NULL;

	int i=0;
	while (attrs[i])
	{
		CASE ("qobject",       etype)
		CASE ("table",         tabname)
		CASE_UNKNOWN;
		i += 2;
	}
	REQUIRE2(etype,tabname, "An SQL table name and an object type!");

	ctx->map = qof_map_new(etype, tabname);
}

static void
qmap_end (gpointer self)
{
	QmapParseCtxt *ctx = self;

	ctx->map_list = g_list_append (ctx->map_list, ctx->map);
	ctx->map = NULL;
}

static DuiParserElementPlugin *
qmap_plugin (QmapParseCtxt *ctx)
{
	DuiParserElementPlugin *plg = g_new0 (DuiParserElementPlugin, 1);

	plg->eltname = "qof-map";
	plg->start_handler = qmap_start;
	plg->end_handler = qmap_end;
	plg->user_data = ctx;
	
	return plg;
}

/* ============================================================== */

QmapParseCtxt *
qmap_new (DuiParser *par)
{
	QmapParseCtxt *ctx = g_new0 (QmapParseCtxt, 1);

	DuiParserElementPlugin *plg = qmap_plugin(ctx);
	dui_parser_register_element_plugin (par, plg);

	plg = qmap_field_plugin(ctx);
	dui_parser_register_element_plugin (par, plg);

	plg = qmap_match_plugin(ctx);
	dui_parser_register_element_plugin (par, plg);

	plg = qmap_version_plugin(ctx);
	dui_parser_register_element_plugin (par, plg);

	return ctx;
}

GList *
qmap_get_maplist (QmapParseCtxt *ctx)
{
	if (!ctx) return NULL;
	return ctx->map_list;
}

/* ======================== END OF FILE ========================= */
/* ============================================================== */
